
import bpy
import json
from pathlib import Path
import os
import re
import time
from pathlib import Path
from mathutils import Vector, geometry
from pprint import pprint as pp
from ..constants import PREFIX, BGCOL, INIT_POS, MODULE_DIR
from ..core import (create_cam, set_collection, get_image_infos_from_object, get_col, 
    import_image_as_gp_reference, create_image_plane, set_collection, get_image, 
    create_empty_image, create_plane_holder, create_plane_driver, reload_bg_list,
    link_nodegroup)


def get_json_infos(json_path) -> tuple((list, tuple)):
    '''return a tuple with (image paths list, [x, y] original psd resolution'''

    import json
    setup_json_path = Path(json_path)
    setup_json = json.loads(setup_json_path.read_text())

    # psd = Path(setup_json['psd']) # path to psd
    # image_size = setup_json['image_size'] # image size (reduced)
    # png_dir = Path(setup_json['png_dir']) # path to png directory
    layers = setup_json['layers'] # dic : name, lbbox, bbox, index, path
    org_image_size = setup_json['org_image_size'] # original PSD dimensions
    img_list = [Path(l['path']) for l in layers] # (png_dir /l['name']).with_suffix('.png')

    return img_list, org_image_size


def setup_bg_camera(image_size, scene=None, cam_type=None):
    '''Create background camera from scene active camera according to background image resolution
    image_size :: 
    cam_type :: wether the BG camera should be set to perspective or orthographic
    '''

    scn = scene or bpy.context.scene

    movie_resolution = scn.render.resolution_x, scn.render.resolution_y
    anim_cam = scn.camera
    anim_cam_scale = 6 # anim_cam.data.ortho_scale
    anim_cam_focale = 50 # anim_cam.data.lens
    cam_type = cam_type or anim_cam.data.type

    ## scale up plane scale and bg_cam ortho_scale according to film res
    bg_factor = (anim_cam_scale * image_size[0]) / movie_resolution[0]
    #print(f'({anim_cam_scale} * {image_size[0]}) / {movie_resolution[0]} = {bg_factor}')
    
    ## scale down the focal according to film res
    bg_focale = (anim_cam_focale * movie_resolution[0]) / image_size[0]
    #print(f'({anim_cam_focale} * {movie_resolution[0]}) / {image_size[0]} = {bg_focale}')

    bg_cam = scn.objects.get('bg_cam')
    if not bg_cam:
        bg_cam = create_cam(name='bg_cam', type=cam_type, size=bg_factor, focale=bg_focale)
        bg_cam['resolution'] = image_size
    
    bg_cam.matrix_world = anim_cam.matrix_world
    
    bg_cam.hide_select = True
    bg_cam.lock_location = bg_cam.lock_rotation = bg_cam.lock_scale = [True]*3

    set_collection(bg_cam, 'Camera')
    set_collection(anim_cam, 'Camera')
    return bg_cam

def import_planes(images, import_type='GREASEPENCIL', mode='REPLACE', image_size=None):
    scn = bpy.context.scene

    existing_backgrounds = [(o, *img_info) for o in scn.objects \
                                if o.get('is_background') and (img_info := get_image_infos_from_object(o)) and o.get('is_background') and o.parent]
        
    existing_backgrounds.sort(key=lambda x: x[0].parent.location.z, reverse=True) # sort by parent (holder) location Z
    
    far_plane = INIT_POS
    if existing_backgrounds:
        far_plane = existing_backgrounds[-1][0].parent.location.z

    ## Ensure bg_cam setup (option to use active camera ?)
    bg_cam = scn.objects.get('bg_cam')
    if not bg_cam:
        if not image_size:
            ## Get image size from first file
            img = bpy.data.images.load(str(images[0]), check_existing=True) 
            image_size = (img.size[0], img.size[1])
        bg_cam = setup_bg_camera(image_size)

    ## Ensure Background collection
    backgrounds = get_col(BGCOL) # Create if needed

    for image_path in images:
        image_path = Path(image_path)
        print(f'Importing {image_path.name}')
        current_holder = None

        file_stem = image_path.stem
        # current_bg = next((o for o if o.parent and get_image_infos_from_object(o)), None)
        current_bg = next((o for o in existing_backgrounds if Path(o[1].filepath) == image_path), None)
        if current_bg:
            current_holder = current_bg.parent
            # TODO store opacity or delete existing objects only after loop
        
        if current_bg and mode == 'SKIP':
            print(f' - SKIP: existing {current_bg.name}')
            continue

        if current_bg and mode == 'REPLACE':
            print(f' - DEL:  existing {current_bg.name}')
            bpy.data.objects.remove(current_bg)

        # Import image
        if import_type == 'GREASEPENCIL':
            bg_img = import_image_as_gp_reference(
                image=str(image_path),
                pack_image=False,
            )
        elif import_type == 'MESH':
            if not (texture_plane_ng := bpy.data.node_groups.get('texture_plane')):
                blend = str(MODULE_DIR / 'texture_plane.blend')
                texture_plane_ng = link_nodegroup(blend, 'texture_plane', link=False)

            bg_img = create_image_plane(image_path, texture_plane_ng)

        elif import_type == 'EMPTY':
            bg_img = create_empty_image(image_path)


        bg_img.name = file_stem + '_texgp'
        bg_img.hide_select = True

        if current_holder:
            bg_img.parent = current_holder
            set_collection(bg_img, current_holder.users_collection[0].name)
            continue      

        # img, opacity = get_image_infos_from_object(bg_img)
        print('bg_img: ', bg_img)
        img = get_image(bg_img)
        # bg_name = clean_image_name(img.name)
        file_stem = file_stem

        # Get create collection from image clean name
        bg_col = get_col(file_stem, parent=backgrounds)

        ## Set in collection
        bg_col.objects.link(bg_img)

        print('img: ', img.name, bg_img.name, bg_img.users_collection)
        
        ## create the holder, parent to camera, set driver and set collection. Could also pass 'fp'
        holder = create_plane_holder(img, name=PREFIX + file_stem, parent=bg_cam, link_in_col=bg_col)
        print('holder: ', holder.name)
        
        create_plane_driver(holder, bg_cam, distance=far_plane)
        bg_img.parent = holder

        far_plane += 2

    